import json
from typing import (
    Annotated,
    Any,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    TypeVar,
    Union,
    cast,
    get_args,
    overload,
)

import pystac
from pydantic import ConfigDict, Field
from pydantic.fields import FieldInfo
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)

from stac_model.base import ModelTask, OmitIfNone
from stac_model.input import ModelInput
from stac_model.output import ModelOutput
from stac_model.runtime import Runtime

T = TypeVar(
    "T",
    pystac.Collection,
    pystac.Item,
    pystac.Asset,  # item_assets.AssetDefinition,
)

SchemaName = Literal["mlm"]
SCHEMA_URI: str = "https://crim-ca.github.io/mlm-extension/v1.1.0/schema.json"
PREFIX = f"{get_args(SchemaName)[0]}:"


def mlm_prefix_adder(field_name: str) -> str:
    return "mlm:" + field_name


class MLModelProperties(Runtime):
    name: str = Field(min_length=1)
    architecture: str = Field(min_length=1)
    tasks: Set[ModelTask]
    input: List[ModelInput]
    output: List[ModelOutput]

    total_parameters: int
    pretrained: Annotated[Optional[bool], OmitIfNone] = Field(default=True)
    pretrained_source: Annotated[Optional[str], OmitIfNone] = None

    model_config = ConfigDict(alias_generator=mlm_prefix_adder, populate_by_name=True, extra="ignore")


class MLModelExtension(
    Generic[T],
    PropertiesExtension,
    # FIXME: resolve typing incompatibility?
    #   'pystac.Asset' does not derive from STACObject
    #   therefore, it technically cannot be used in 'ExtensionManagementMixin[T]'
    #   however, this makes our extension definition much easier and avoids lots of code duplication
    ExtensionManagementMixin[  # type: ignore[type-var]
        Union[
            pystac.Collection,
            pystac.Item,
            pystac.Asset,
        ]
    ],
):
    @property
    def name(self) -> SchemaName:
        return cast(SchemaName, get_args(SchemaName)[0])

    def apply(
        self,
        properties: Union[MLModelProperties, dict[str, Any]],
    ) -> None:
        """
        Applies Machine Learning Model Extension properties to the extended :mod:`~pystac` object.
        """
        if isinstance(properties, dict):
            properties = MLModelProperties(**properties)
        data_json = json.loads(properties.model_dump_json(by_alias=True))
        for prop, val in data_json.items():
            self._set_property(prop, val)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @overload
    @classmethod
    def ext(cls, obj: pystac.Asset, add_if_missing: bool = False) -> "AssetMLModelExtension": ...

    @overload
    @classmethod
    def ext(cls, obj: pystac.Item, add_if_missing: bool = False) -> "ItemMLModelExtension": ...

    @overload
    @classmethod
    def ext(cls, obj: pystac.Collection, add_if_missing: bool = False) -> "CollectionMLModelExtension": ...

    # @overload
    # @classmethod
    # def ext(cls, obj: item_assets.AssetDefinition, add_if_missing: bool = False) -> "ItemAssetsMLModelExtension":
    #     ...

    @classmethod
    def ext(
        cls,
        obj: Union[pystac.Collection, pystac.Item, pystac.Asset],  # item_assets.AssetDefinition
        add_if_missing: bool = False,
    ) -> Union[
        "CollectionMLModelExtension",
        "ItemMLModelExtension",
        "AssetMLModelExtension",
    ]:
        """
        Extends the given STAC Object with properties from the :stac-ext:`Machine Learning Model Extension <mlm>`.

        This extension can be applied to instances of :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Args:
            obj: STAC Object to extend with the MLM extension fields.
            add_if_missing: Add the MLM extension schema URI to the object if not already in `stac_extensions`.

        Returns:
            Extended object.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return CollectionMLModelExtension(obj)
        elif isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return ItemMLModelExtension(obj)
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return AssetMLModelExtension(obj)
        # elif isinstance(obj, item_assets.AssetDefinition):
        #     cls.ensure_owner_has_extension(obj, add_if_missing)
        #     return ItemAssetsMLModelExtension(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(cls, obj: pystac.Collection, add_if_missing: bool = False) -> "SummariesMLModelExtension":
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesMLModelExtension(obj)


class SummariesMLModelExtension(SummariesExtension):
    """
    Summaries annotated with the Machine Learning Model Extension.

    A concrete implementation of :class:`~SummariesExtension` that extends
    the ``summaries`` field of a :class:`~pystac.Collection` to include properties
    defined in the :stac-ext:`Machine Learning Model <mlm>`.
    """

    def _check_mlm_property(self, prop: str) -> FieldInfo:
        try:
            return MLModelProperties.model_fields[prop]
        except KeyError as err:
            raise AttributeError(f"Name '{prop}' is not a valid MLM property.") from err

    def _validate_mlm_property(self, prop: str, summaries: list[Any]) -> None:
        # ignore mypy issue when combined with Annotated
        #   - https://github.com/pydantic/pydantic/issues/6713
        #   - https://github.com/pydantic/pydantic/issues/5190
        model = MLModelProperties.model_construct()  # type: ignore[call-arg]
        validator = MLModelProperties.__pydantic_validator__
        for value in summaries:
            validator.validate_assignment(model, prop, value)

    def get_mlm_property(self, prop: str) -> Optional[list[Any]]:
        self._check_mlm_property(prop)
        return self.summaries.get_list(prop)

    def set_mlm_property(self, prop: str, summaries: list[Any]) -> None:
        self._check_mlm_property(prop)
        self._validate_mlm_property(prop, summaries)
        self._set_summary(prop, summaries)

    def __getattr__(self, prop):
        return self.get_mlm_property(prop)

    def __setattr__(self, prop, value):
        self.set_mlm_property(prop, value)


class ItemMLModelExtension(MLModelExtension[pystac.Item]):
    """
    Item annotated with the Machine Learning Model Extension.

    A concrete implementation of :class:`MLModelExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to
    include properties defined in the :stac-ext:`Machine Learning Model
    Extension <mlm>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`MLModelExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemMLModelExtension Item id={self.item.id}>"


# class ItemAssetsMLModelExtension(MLModelExtension[item_assets.AssetDefinition]):
#     properties: dict[str, Any]
#     asset_defn: item_assets.AssetDefinition
#
#     def __init__(self, item_asset: item_assets.AssetDefinition):
#         self.asset_defn = item_asset
#         self.properties = item_asset.properties


class AssetMLModelExtension(MLModelExtension[pystac.Asset]):
    """
    Asset annotated with the Machine Learning Model Extension.

    A concrete implementation of :class:`MLModelExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include
    properties defined in the :stac-ext:`Machine Learning Model
    Extension <mlm>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`MLModelExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Optional[Iterable[dict[str, Any]]] = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetMLModelExtension Asset href={self.asset_href}>"


class CollectionMLModelExtension(MLModelExtension[pystac.Collection]):
    def __init__(self, collection: pystac.Collection):
        self.collection = collection


# __all__ = [
#     "MLModelExtension",
#     "ModelInput",
#     "InputArray",
#     "Band",
#     "Statistics",
#     "ModelOutput",
#     "Asset",
#     "Runtime",
#     "Container",
#     "Asset",
# ]



"""
A PydanticV2/PySTAC validation and serialization library for the STAC Machine Learning Model Extension.
"""

from importlib import metadata

try:
    __version__ = metadata.version("stac-model")
except metadata.PackageNotFoundError:
    __version__ = "unknown"



import json

import typer
from rich.console import Console

from stac_model import __version__
from stac_model.examples import eurosat_resnet
from stac_model.schema import ItemMLModelExtension

app = typer.Typer(
    name="stac-model",
    help="A PydanticV2 validation and serialization library for the STAC Machine Learning Model Extension",
    add_completion=False,
)
console = Console()


def version_callback(print_version: bool) -> None:
    """Print the version of the package."""
    if print_version:
        console.print(f"[yellow]stac-model[/] version: [bold blue]{__version__}[/]")
        raise typer.Exit()


@app.command(name="")
def main(
    print_version: bool = typer.Option(
        None,
        "-v",
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Prints the version of the stac-model package.",
    ),
) -> ItemMLModelExtension:
    """Generate example spec."""
    ml_model_meta = eurosat_resnet()
    with open("example.json", "w") as json_file:
        json.dump(ml_model_meta.item.to_dict(), json_file, indent=4)
    print("Example model metadata written to ./example.json.")
    return ml_model_meta


if __name__ == "__main__":
    app()



from enum import Enum
from typing import Annotated, Literal, Optional, Union

from pydantic import AliasChoices, Field

from stac_model.base import MLMBaseModel, OmitIfNone


class AcceleratorEnum(str, Enum):
    amd64 = "amd64"
    cuda = "cuda"
    xla = "xla"
    amd_rocm = "amd-rocm"
    intel_ipex_cpu = "intel-ipex-cpu"
    intel_ipex_gpu = "intel-ipex-gpu"
    macos_arm = "macos-arm"

    def __str__(self):
        return self.value


AcceleratorName = Literal[
    "amd64",
    "cuda",
    "xla",
    "amd-rocm",
    "intel-ipex-cpu",
    "intel-ipex-gpu",
    "macos-arm",
]

AcceleratorType = Union[AcceleratorName, AcceleratorEnum]


class Runtime(MLMBaseModel):
    framework: Annotated[Optional[str], OmitIfNone] = Field(default=None)
    framework_version: Annotated[Optional[str], OmitIfNone] = Field(default=None)
    file_size: Annotated[Optional[int], OmitIfNone] = Field(
        alias="file:size",
        validation_alias=AliasChoices("file_size", "file:size"),
        default=None,
    )
    memory_size: Annotated[Optional[int], OmitIfNone] = Field(default=None)
    batch_size_suggestion: Annotated[Optional[int], OmitIfNone] = Field(default=None)

    accelerator: Optional[AcceleratorType] = Field(default=None)
    accelerator_constrained: bool = Field(default=False)
    accelerator_summary: Annotated[Optional[str], OmitIfNone] = Field(default=None)
    accelerator_count: Annotated[Optional[int], OmitIfNone] = Field(default=None, minimum=1)



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Literal, TypeAlias, Union

from pydantic import BaseModel, ConfigDict, model_serializer

Number: TypeAlias = Union[int, float]
JSON: TypeAlias = Union[
    Dict[str, "JSON"],
    List["JSON"],
    Number,
    bool,
    str,
    None,
]


@dataclass
class _OmitIfNone:
    pass


OmitIfNone = _OmitIfNone()


class MLMBaseModel(BaseModel):
    """
    Allows wrapping any field with an annotation to drop it entirely if unset.

    ```python
    field: Annotated[Optional[<desiredType>], OmitIfNone] = None
    # or
    field: Annotated[Optional[<desiredType>], OmitIfNone] = Field(default=None)
    ```

    Since `OmitIfNone` implies that the value could be `None` (even though it would be dropped),
    the `Optional` annotation must be specified to corresponding typings to avoid `mypy` lint issues.

    It is important to use `MLMBaseModel`, otherwise the serializer will not be called and applied.

    Reference: https://github.com/pydantic/pydantic/discussions/5461#discussioncomment-7503283
    """

    @model_serializer
    def model_serialize(self):
        omit_if_none_fields = {
            key: field
            for key, field in self.model_fields.items()
            if any(isinstance(m, _OmitIfNone) for m in field.metadata)
        }
        values = {
            self.__fields__[key].alias or key: val  # use the alias if specified
            for key, val in self
            if key not in omit_if_none_fields or val is not None
        }
        return values

    model_config = ConfigDict(
        populate_by_name=True,
    )


DataType: TypeAlias = Literal[
    "uint8",
    "uint16",
    "uint32",
    "uint64",
    "int8",
    "int16",
    "int32",
    "int64",
    "float16",
    "float32",
    "float64",
    "cint16",
    "cint32",
    "cfloat32",
    "cfloat64",
    "other",
]


class TaskEnum(str, Enum):
    REGRESSION = "regression"
    CLASSIFICATION = "classification"
    SCENE_CLASSIFICATION = "scene-classification"
    DETECTION = "detection"
    OBJECT_DETECTION = "object-detection"
    SEGMENTATION = "segmentation"
    SEMANTIC_SEGMENTATION = "semantic-segmentation"
    INSTANCE_SEGMENTATION = "instance-segmentation"
    PANOPTIC_SEGMENTATION = "panoptic-segmentation"
    SIMILARITY_SEARCH = "similarity-search"
    GENERATIVE = "generative"
    IMAGE_CAPTIONING = "image-captioning"
    SUPER_RESOLUTION = "super-resolution"


ModelTaskNames: TypeAlias = Literal[
    "regression",
    "classification",
    "scene-classification",
    "detection",
    "object-detection",
    "segmentation",
    "semantic-segmentation",
    "instance-segmentation",
    "panoptic-segmentation",
    "similarity-search",
    "generative",
    "image-captioning",
    "super-resolution",
]


ModelTask = Union[ModelTaskNames, TaskEnum]


class ProcessingExpression(BaseModel):
    # FIXME: should use 'pystac' reference, but 'processing' extension is not implemented yet!
    format: str
    expression: Any



from typing import Annotated, Any, Dict, List, Optional, Set, Union, cast

from pydantic import AliasChoices, ConfigDict, Field, model_serializer
from pystac.extensions.classification import Classification

from stac_model.base import DataType, MLMBaseModel, ModelTask, OmitIfNone, ProcessingExpression


class ModelResult(MLMBaseModel):
    shape: List[Union[int, float]] = Field(..., min_items=1)
    dim_order: List[str] = Field(..., min_items=1)
    data_type: DataType


# MLMClassification: TypeAlias = Annotated[
#     Classification,
#     PlainSerializer(
#         lambda x: x.to_dict(),
#         when_used="json",
#         return_type=TypedDict(
#             "Classification",
#             {
#                 "value": int,
#                 "name": str,
#                 "description": NotRequired[str],
#                 "color_hint": NotRequired[str],
#             }
#         )
#     )
# ]


class MLMClassification(MLMBaseModel, Classification):
    @model_serializer()
    def model_dump(self, *_: Any, **__: Any) -> Dict[str, Any]:
        return self.to_dict()  # type: ignore[call-arg]

    def __init__(
        self,
        value: int,
        description: Optional[str] = None,
        name: Optional[str] = None,
        color_hint: Optional[str] = None,
    ) -> None:
        Classification.__init__(self, {})
        if not name and not description:
            raise ValueError("Class name or description is required!")
        self.apply(
            value=value,
            name=name or description,
            description=cast(str, description or name),
            color_hint=color_hint,
        )

    def __hash__(self) -> int:
        return sum(map(hash, self.to_dict().items()))

    def __setattr__(self, key: str, value: Any) -> None:
        if key == "properties":
            Classification.__setattr__(self, key, value)
        else:
            MLMBaseModel.__setattr__(self, key, value)

    model_config = ConfigDict(
        populate_by_name=True,
        arbitrary_types_allowed=True,
    )


# class ClassObject(BaseModel):
#     value: int
#     name: str
#     description: Optional[str] = None
#     title: Optional[str] = None
#     color_hint: Optional[str] = None
#     nodata: Optional[bool] = False


class ModelOutput(MLMBaseModel):
    name: str
    tasks: Set[ModelTask]
    result: ModelResult

    # NOTE:
    #   Although it is preferable to have 'Set' to avoid duplicate,
    #   it is more important to keep the order in this case,
    #   which we would lose with 'Set'.
    #   We also get some unhashable errors with 'Set', although 'MLMClassification' implements '__hash__'.
    classes: Annotated[List[MLMClassification], OmitIfNone] = Field(
        alias="classification:classes",
        validation_alias=AliasChoices("classification:classes", "classification_classes", "classes"),
    )
    post_processing_function: Optional[ProcessingExpression] = None

    model_config = ConfigDict(
        populate_by_name=True,
    )



from typing import Annotated, List, Literal, Optional, TypeAlias, Union

from pydantic import Field

from stac_model.base import DataType, MLMBaseModel, Number, OmitIfNone, ProcessingExpression


class InputStructure(MLMBaseModel):
    shape: List[Union[int, float]] = Field(min_items=1)
    dim_order: List[str] = Field(min_items=1)
    data_type: DataType


class MLMStatistic(MLMBaseModel):  # FIXME: add 'Statistics' dep from raster extension (cases required to be triggered)
    minimum: Annotated[Optional[Number], OmitIfNone] = None
    maximum: Annotated[Optional[Number], OmitIfNone] = None
    mean: Annotated[Optional[Number], OmitIfNone] = None
    stddev: Annotated[Optional[Number], OmitIfNone] = None
    count: Annotated[Optional[int], OmitIfNone] = None
    valid_percent: Annotated[Optional[Number], OmitIfNone] = None


NormalizeType: TypeAlias = Optional[
    Literal[
        "min-max",
        "z-score",
        "l1",
        "l2",
        "l2sqr",
        "hamming",
        "hamming2",
        "type-mask",
        "relative",
        "inf"
    ]
]

ResizeType: TypeAlias = Optional[
    Literal[
        "crop",
        "pad",
        "interpolation-nearest",
        "interpolation-linear",
        "interpolation-cubic",
        "interpolation-area",
        "interpolation-lanczos4",
        "interpolation-max",
        "wrap-fill-outliers",
        "wrap-inverse-map",
    ]
]


class ModelInput(MLMBaseModel):
    name: str
    bands: List[str]  # order is critical here (same index as dim shape), allow duplicate if the model needs it somehow
    input: InputStructure
    norm_by_channel: Annotated[Optional[bool], OmitIfNone] = None
    norm_type: Annotated[Optional[NormalizeType], OmitIfNone] = None
    norm_clip: Annotated[Optional[List[Union[float, int]]], OmitIfNone] = None
    resize_type: Annotated[Optional[ResizeType], OmitIfNone] = None
    statistics: Annotated[Optional[List[MLMStatistic]], OmitIfNone] = None
    pre_processing_function: Optional[ProcessingExpression] = None



from typing import cast

import pystac
import shapely
from dateutil.parser import parse as parse_dt
from pystac.extensions.file import FileExtension

from stac_model.base import ProcessingExpression
from stac_model.input import InputStructure, MLMStatistic, ModelInput
from stac_model.output import MLMClassification, ModelOutput, ModelResult
from stac_model.schema import ItemMLModelExtension, MLModelExtension, MLModelProperties


def eurosat_resnet() -> ItemMLModelExtension:
    input_struct = InputStructure(
        shape=[-1, 13, 64, 64],
        dim_order=["batch", "channel", "height", "width"],
        data_type="float32",
    )
    band_names = [
        "B01",
        "B02",
        "B03",
        "B04",
        "B05",
        "B06",
        "B07",
        "B08",
        "B8A",
        "B09",
        "B10",
        "B11",
        "B12",
    ]
    stats_mean = [
        1354.40546513,
        1118.24399958,
        1042.92983953,
        947.62620298,
        1199.47283961,
        1999.79090914,
        2369.22292565,
        2296.82608323,
        732.08340178,
        12.11327804,
        1819.01027855,
        1118.92391149,
        2594.14080798,
    ]
    stats_stddev = [
        245.71762908,
        333.00778264,
        395.09249139,
        593.75055589,
        566.4170017,
        861.18399006,
        1086.63139075,
        1117.98170791,
        404.91978886,
        4.77584468,
        1002.58768311,
        761.30323499,
        1231.58581042,
    ]
    stats = [
        MLMStatistic(mean=mean, stddev=stddev)
        for mean, stddev in zip(stats_mean, stats_stddev)
    ]
    model_input = ModelInput(
        name="13 Band Sentinel-2 Batch",
        bands=band_names,
        input=input_struct,
        norm_by_channel=True,
        norm_type="z-score",
        resize_type=None,
        statistics=stats,
        pre_processing_function=ProcessingExpression(
            format="python",
            expression="torchgeo.datamodules.eurosat.EuroSATDataModule.collate_fn",
        ),  # noqa: E501
    )
    result_struct = ModelResult(
        shape=[-1, 10],
        dim_order=["batch", "class"],
        data_type="float32"
    )
    class_map = {
        "Annual Crop": 0,
        "Forest": 1,
        "Herbaceous Vegetation": 2,
        "Highway": 3,
        "Industrial Buildings": 4,
        "Pasture": 5,
        "Permanent Crop": 6,
        "Residential Buildings": 7,
        "River": 8,
        "SeaLake": 9,
    }
    class_objects = [
        MLMClassification(value=class_value, name=class_name)
        for class_name, class_value in class_map.items()
    ]
    model_output = ModelOutput(
        name="classification",
        tasks={"classification"},
        classes=class_objects,
        result=result_struct,
        post_processing_function=None,
    )
    assets = {
        "model": pystac.Asset(
            title="Pytorch weights checkpoint",
            description=(
                "A Resnet-18 classification model trained on normalized Sentinel-2 "
                "imagery with Eurosat landcover labels with torchgeo."
            ),
            href="https://huggingface.co/torchgeo/resnet18_sentinel2_all_moco/resolve/main/resnet18_sentinel2_all_moco-59bfdff9.pth",
            media_type="application/octet-stream; application=pytorch",
            roles=[
                "mlm:model",
                "mlm:weights",
                "data"
            ]
        ),
        "source_code": pystac.Asset(
            title="Model implementation.",
            description="Source code to run the model.",
            href="https://github.com/microsoft/torchgeo/blob/61efd2e2c4df7ebe3bd03002ebbaeaa3cfe9885a/torchgeo/models/resnet.py#L207",
            media_type="text/x-python",
            roles=[
                "mlm:model",
                "code"
            ]
        )
    }

    ml_model_size = 43000000
    ml_model_meta = MLModelProperties(
        name="Resnet-18 Sentinel-2 ALL MOCO",
        architecture="ResNet-18",
        tasks={"classification"},
        framework="pytorch",
        framework_version="2.1.2+cu121",
        accelerator="cuda",
        accelerator_constrained=False,
        accelerator_summary="Unknown",
        file_size=ml_model_size,
        memory_size=1,
        pretrained=True,
        pretrained_source="EuroSat Sentinel-2",
        total_parameters=11_700_000,
        input=[model_input],
        output=[model_output],
    )
    # TODO, this can't be serialized but pystac.item calls for a datetime
    # in docs. start_datetime=datetime.strptime("1900-01-01", "%Y-%m-%d")
    # Is this a problem that we don't do date validation if we supply as str?
    start_datetime_str = "1900-01-01"
    end_datetime_str = "9999-01-01"  # cannot be None, invalid against STAC Core!
    start_datetime = parse_dt(start_datetime_str).isoformat() + "Z"
    end_datetime = parse_dt(end_datetime_str).isoformat() + "Z"
    bbox = [
        -7.882190080512502,
        37.13739173208318,
        27.911651652899923,
        58.21798141355221
    ]
    geometry = shapely.geometry.Polygon.from_bounds(*bbox).__geo_interface__
    item_name = "item_basic"
    col_name = "ml-model-examples"
    item = pystac.Item(
        id=item_name,
        collection=col_name,
        geometry=geometry,
        bbox=bbox,
        datetime=None,
        properties={
            "start_datetime": start_datetime,
            "end_datetime": end_datetime,
            "description": (
                "Sourced from torchgeo python library, identifier is ResNet18_Weights.SENTINEL2_ALL_MOCO"
            ),
        },
        assets=assets,
    )

    # note: cannot use 'item.add_derived_from' since it expects a 'Item' object, but we refer to a 'Collection' here
    # item.add_derived_from("https://earth-search.aws.element84.com/v1/collections/sentinel-2-l2a")
    item.add_link(
        pystac.Link(
            target="https://earth-search.aws.element84.com/v1/collections/sentinel-2-l2a",
            rel=pystac.RelType.DERIVED_FROM,
            media_type=pystac.MediaType.JSON,
        )
    )

    # define more link references
    col = pystac.Collection(
        id=col_name,
        title="Machine Learning Model examples",
        description="Collection of items contained in the Machine Learning Model examples.",
        extent=pystac.Extent(
            temporal=pystac.TemporalExtent([[parse_dt(start_datetime), parse_dt(end_datetime)]]),
            spatial=pystac.SpatialExtent([bbox]),
        )
    )
    col.set_self_href("./examples/collection.json")
    col.add_item(item)
    item.set_self_href(f"./examples/{item_name}.json")

    model_asset = cast(
        FileExtension[pystac.Asset],
        pystac.extensions.file.FileExtension.ext(assets["model"], add_if_missing=True)
    )
    model_asset.apply(size=ml_model_size)

    item_mlm = MLModelExtension.ext(item, add_if_missing=True)
    item_mlm.apply(ml_model_meta.model_dump(by_alias=True, exclude_unset=True, exclude_defaults=True))
    return item_mlm
